/*
m_spectrip - An additive oscillator

DESCRIPTION
An additive oscillator, allowing for different (equally spaced) harmonic
spectra and different amplitude curves for the harmonics. It also comes
with a stereo spread (spreading alternating harmonics to the left and right
channel) and a comb filter, with its frequency always related to the pitch
of the current note.

SYNTAX
aLeft, aRight m_spectrip kamp, kcps, kSpread, kMorph, kSpacing, kBright, \
	kResonanceVolume, kResonanceRatio, kResonanceFeedback

INITIALIZATION

PERFORMANCE
aLeft, aRight - Left and right output from the oscillator. Depending on the
	spread parameter this can be the same, up to full separation of alternating
	partials
kamp - Amplitude of the oscillator (0 to 1)
kcps - Pitch of the oscillator in Hz
kSpread - Stereo spread of the partials (0 to 1), where 0 is mono and
	1 is full stereo
kMorph - Spectrum tilt (0 to 3), characteristic of the spectrum, there are
	four different curves:
	0: 1/n^2 (triangle with kSpacing of 2)
	1: 1/n (sawtooth with all harmonics)
	2: cosine
	3: cubic curve
kSpacing - Spacing or ratio between the harmonics, (.5 to 5)
	1 marks the full harmonic spectrum (a sawtooth), 2 (every second harmonic
	counted from 1, a square wave)
kBright - Brightness (between 0 and 1), this will fade out higher harmonics
kResonanceVolume - Volume of the buyilt-in comb filter (can give a nice
	flanger-like effect, (between 0 and 1)
kResonanceRatio - Frequency ratio of the comb filter (.75 to 5) times kcps
	(i.e. the frequency of the note played)
kResonanceFeedback - Feedback of the comb filter (0 to 1)

CREDITS
Author: Jeanette C.
*/

opcode m_spectrip, aa, kkkkkkkkk
	kamp, kcps, kSpread, kMorph, kSpacing, kBright, \
	kResonanceVolume, kResonanceRatio, kResonanceFeedback xin

	; Limit kMorph for safety
	kMorph limit kMorph, 0, 3
	kCombVol init 0 ; actual volume of the comb filter

	kSpacingChanged init 1 ; when kSpacing changes spec-tables need
		; re-evaluation

	; Create the necessary ftable
	ihyper1 ftgenonce 0, 0, 128, -23, "hyper1.ftable" ; 1/x spectrum
	ihyper2 ftgenonce 0, 0, 128, -23, "hyper2.ftable" ; 1/x^2 spectrum
	icosbase ftgenonce 0, 0, 128, 9, 1, .001, 90 ; Cosine spectrum
	icos ftgenonce 0, 0, 128, -24, icosbase, 1/128, 1/32
	iline ftgenonce 0, 0, 128, -6, 1/32, 128, 1/128 ; equal volume spectrum

	; ftables for left and right amplitudes
	ihyper11 ftgenonce 200, 0, 64, 7, 0, 64, 0
	ihyper12 ftgenonce 201, 0, 64, 7, 0, 64, 0
	ihyper21 ftgenonce 202, 0, 64, 7, 0, 64, 0
	ihyper22 ftgenonce 203, 0, 64, 7, 0, 64, 0
	iline1 ftgenonce 204, 0, 64, 7, 0, 64, 0
	iline2 ftgenonce 205, 0, 64, 7, 0, 64, 0
	icos1 ftgenonce 206, 0, 64, 7, 0, 64, 0
	icos2 ftgenonce 207, 0, 64, 7, 0, 64, 0

	; Tables holding morph indices for left and right parts
	imorph1 ftgenonce 208, 0, 4, -2, ihyper21, ihyper11, icos1, iline1
	imorph2 ftgenonce 209, 0, 4, -2, ihyper22, ihyper12, icos2, iline2
	iamps1 ftgenonce 210, 0, 64, 7, 0, 64, 0 ; spectrum amplitudes for part1
	iamps2 ftgenonce 211, 0, 64, 7, 0, 64, 0 ; spectrum amplitudes for part2
	ispec1 ftgenonce 212, 0, 64, 7, 1, 64, 127 ; frequency ratios for part1
	ispec2 ftgenonce 213, 0, 64, 7, 2, 64, 128 ; frequency ratios for part2
	isine ftgenonce 0, 0, 32768, 10, 1 ; sine wave for audio generation

	; Fill amplitude tables for left and right on init or spacing change
	kInitialised init 0 ; set to 1 after first fill
	if (kInitialised == 0) then
		kSpacing limit kSpacing, .5, 5
		kHarmonic = 0
		kIndex = 0
		while (kIndex <64) do
			if ((kHarmonic + 1) <= 128) then
				kvalue tab kHarmonic, ihyper1
				tabw kvalue, kIndex, ihyper11
				kvalue tab (kHarmonic + 1), ihyper1
				tabw kvalue, kIndex, ihyper12
				kvalue tab kHarmonic, ihyper2
				tabw kvalue, kIndex, ihyper21
				kvalue tab (kHarmonic + 1), ihyper2
				tabw kvalue, kIndex, ihyper22
				kvalue tab kHarmonic, icos
				tabw kvalue, kIndex, icos1
				kvalue tab (kHarmonic + 1), icos
				tabw kvalue, kIndex, icos2
				kvalue tab kHarmonic, iline
				tabw kvalue, kIndex, iline1
				kvalue tab (kHarmonic + 1), iline
				tabw kvalue, kIndex, iline2
			else ; Harmonic above 128, will have 0 volume
				tabw 0, kIndex, ihyper11
				tabw 0, kIndex, ihyper21
				tabw 0, kIndex, ihyper12
				tabw 0, kIndex, ihyper22
				tabw 0, kIndex, iline1
				tabw 0, kIndex, iline2
				tabw 0, kIndex, icos1
				tabw 0, kIndex, icos2
			endif

			kIndex += 1
			kHarmonic += 2
		od
		kInitialised = 1
	endif

	if (kSpacingChanged >0) then
		kIndex = 0
		kHarmonic = 0
		while (kIndex <64) && (kHarmonic < 128) do
			; Generate the harmonic spectrum
			kvalue = kHarmonic + 1
			tabw kvalue, kIndex, ispec1
			kvalue += kSpacing
			tabw kvalue, kIndex, ispec2
			kIndex += 1
			kHarmonic += kSpacing
		od
		kSpacingChanged = 0
	endif
	kSpacingChanged changed kSpacing

	; Calculate the amplitudes
	kdrift oscil .00001, 1, isine
	ftmorf kMorph+kdrift, imorph1, iamps1
	ftmorf kMorph+kdrift, imorph2, iamps2

	; Limit the number of audible partials based on brightness
	kPartNum = limit(kBright, 0, 1) * 128
	kCurHarm = 128
	kIndex = 63
	while (kPartNum < kCurHarm) && (kIndex >0) do
	; Check first partial signal
		kCurHarm tab kIndex, ispec1
		if (kCurHarm > kPartNum) then
			tabw 0, kIndex, iamps1
		elseif (int(kCurHarm) == int(kPartNum)) then
			kCurAmp tab kIndex, iamps1
			tabw (kCurAmp * frac(kPartNum)), kIndex, iamps1
		endif

		; Optionally switch partials off abruptly
		;if (kCurHarm >= kPartNum) then
		;	tabw 0, kIndex, iamps1
		;endif

		; Check second partial signal
		kCurHarm tab kIndex, ispec2
		if (kCurHarm > kPartNum) then
			tabw 0, kIndex, iamps2
		elseif (int(kCurHarm) == int(kPartNum)) then
			kCurAmp tab kIndex, iamps2
			tabw (kCurAmp * frac(kPartNum)), kIndex, iamps2
		endif

		; Optionally switch partials off abruptly
		;if (kCurHarm >= kPartNum) then
		;	tabw 0, kIndex, iamps2
		;endif

		kIndex -= 1
	od

	; Produce the additive signals
	aPart1 adsynt kamp, kcps, isine, ispec1, iamps1, 64
	aPart2 adsynt kamp, kcps, isine, ispec2, iamps2, 64

	; Comb filter the two partial signals
	if (kResonanceVolume >0) then
		kCombVol limit (kResonanceVolume * .25), 0, .25
		kRatio limit kResonanceRatio, .75, 5
		kCombFreq = 1/(kcps * kRatio)
		kRevTime = limit(kResonanceFeedback, 0, 1) * .25
		aComb1 vcomb aPart1*kCombVol, kRevTime, kCombFreq, .03
		aComb2 vcomb aPart2*kCombVol, kRevTime, kCombFreq, .03
	else
		aComb1 = 0
		aComb2 = 0
	endif

	; Stereo spread/panning
	if (kSpread < 1) then
		kGain = .3
	else
		kGain = .6
	endif
	kPan limit (kSpread / 2), 0, .5 ; Tame the parameter for safety :)
	kSameVol = .5 + kPan
	kOppositeVol = .5 - kPan
	aLeft = ((aPart1 + aComb1) * kSameVol + (aPart2 + aComb2) * kOppositeVol) * kGain * (.9 - kCombVol - kRevTime)
	aRight = ((aPart2 + aComb2) * kSameVol + (aPart1 + aComb1) * kOppositeVol) * kGain * (.9 - kCombVol - kRevTime)

	; Output the audio
	xout aLeft, aRight
endop
